from .ae import *
from .maae import *
from .mwae import *
from .mcvae import *
from .mvae import *
from .mmvae import *
from .mvtcae import *
from .dvcca import *
from .me_mvae import *
from .jmvae import *
from .mopoevae import *
from .mmJSD import *
from .weighted_mvae import *
from .DMVAE import *
from .weighted_DMVAE import *
from .mmvaeplus import *
from .dccae import *
__all__ = ["AE", "mAAE", "mWAE", "mcVAE", "mVAE", "JMVAE", "me_mVAE", "mmVAE", "mvtCAE", "DVCCA", "MoPoEVAE", "mmJSD", "weighted_mVAE",
         "DMVAE", "weighted_DMVAE", "mmVAEPlus", "DCCAE"]
classes = __all__ 

